""" This module implemented for Hepsiburada Data Science Team to interact with ElasticSearch.

Raises:
    NotImplementedError: _description_
    RuntimeError: _description_
    RuntimeError: _description_
    RuntimeError: _description_

Returns:
    _type_: _description_
"""

import os
import time

import pandas as pd
from colorama import Fore, init
from elasticsearch import Elasticsearch, RequestError, helpers

init(autoreset=True)


class ElasticOperations(Elasticsearch):
    """This class is created to enable Hepsiburada Data Science to communicate with ElasticSearch"""

    _implemented_return_methods = ["raw", "hits", "total", "value", None]

    def __init__(self, **kwargs) -> None:

        self.ip = kwargs.get("ip")
        self.port = kwargs.get("port")
        self.batch_length = 30_000
        self.client = self.create_client(self.ip, self.port)

    def create_client(self, ip, port):
        """Creates a elasticsearch client to run operations."""
        self.client = Elasticsearch([ip], port=port)
        return self.client

    def create_index(self, indexName):
        """Creates an index."""
        if self.client.indices.exists(index=indexName):
            print(Fore.YELLOW, f"Index {indexName} already exists.")
        else:
            self.client.indices.create(index=indexName)
            print(Fore.GREEN + f"Index {indexName} is created.")

    def delete_index(self, indexName):
        """Deletes the index."""
        if not self.client.indices.exists(index=indexName):
            print(Fore.YELLOW + f"Index {indexName} does not exist.")
        else:
            self.adjust_readonly(indexName, False)
            self.client.indices.delete(index=indexName)
            print(Fore.GREEN + f"Index {indexName} is deleted.")

    def get_count(self, indexName):
        """Returns the count of the given index."""

        self.client.indices.refresh(indexName)
        count = int(
            self.client.cat.count(indexName, params={"format": "json"})[0][
                "count"
            ]
        )
        return count + 1

    def set_max_result_window(self, indexName):
        """Sets the max_result_window of the index to assign readonly or nonreadonly."""
        count = self.get_count(indexName)

        label = {
            "index.blocks.read_only_allow_delete": "false",
            "index.max_result_window": count,
        }

        self.client.indices.put_settings(label, index=indexName)

    def check_read_only(self, indexName):
        """Checks the given index is read only or not"""

        self.client.indices.refresh(index=indexName)
        result = self.client.indices.get(index=indexName)[indexName]

        try:
            is_read_only = result["settings"]["index"]["blocks"]["read_only"]

            if is_read_only == "true":
                return True
            else:
                return False
        except KeyError:
            return None

    def nonreadonly(self, indexName):
        """Converts index from read only to non-read only"""

        self.set_max_result_window(indexName)

        is_read_only = self.check_read_only(indexName)

        if is_read_only is False:
            print(Fore.YELLOW + f"{indexName} was already not read only.")

        label = {"index.blocks.read_only": "false"}

        self.client.indices.put_settings(label, index=indexName)

    def readonly(self, indexName):
        """Converts index from read only to non-read only"""

        self.set_max_result_window(indexName)

        is_read_only = self.check_read_only(indexName)

        if is_read_only:
            print(Fore.YELLOW + f"{indexName} was already read only.")

        label = {"index.blocks.read_only": "true"}

        self.client.indices.put_settings(label, index=indexName)

    def adjust_readonly(self, indexName, read_only: bool):
        """Converts given index from read only to non-read only or vice versa."""

        self.set_max_result_window(indexName)

        is_read_only = self.check_read_only(indexName)

        if is_read_only and read_only:
            print(Fore.YELLOW + f"{indexName} was already read only.")

        elif not read_only and is_read_only is False:
            print(Fore.YELLOW + f"{indexName} was already not read only.")

        if read_only:
            label = {"index.blocks.read_only": "true"}
        else:
            label = {"index.blocks.read_only": "false"}

        self.client.indices.put_settings(label, index=indexName)

    def recreate_index(self, indexName):
        """Deletes and recreates the assigned index."""
        if not self.client.indices.exists(index=indexName):
            print(Fore.YELLOW + f"Index {indexName} does not exist.")
        else:
            self.adjust_readonly(indexName, False)
            self.client.indices.delete(index=indexName, ignore=[400, 404])
            self.client.indices.create(index=indexName)
            print(Fore.GREEN + f"Index {indexName} is recreated.")

    def send_data_to_elastic(self, indexName, data, timeout=30, verbose=False):
        """Uploads data batch by batch to elastic."""
        self.adjust_readonly(indexName, False)
        start_time = time.time()

        iteration_count = len(data) // self.batch_length + 1
        for idx in range(iteration_count):
            batch = data[
                idx * self.batch_length : (idx + 1) * self.batch_length
            ]
            helpers.bulk(
                self.client,
                batch,
                index=indexName,
                doc_type="_doc",
                request_timeout=timeout,
            )
            if verbose:
                print(
                    Fore.GREEN
                    + f"Batch {idx} is completed at {time.time() - start_time}"
                )
                start_time = time.time()

        self.adjust_readonly(indexName, True)

    def get_data_from_elastic(self, indexName):
        """Downloads data from given index."""

        count = self.get_count(indexName)
        size = count if count <= self.batch_length else self.batch_length
        count -= self.batch_length
        res = self.client.search(
            index=indexName,
            body={
                "query": {"match_all": {}},
                "sort": [{"_id": {"order": "asc"}}],
            },
            size=size,
        )

        result_data = [document["_source"] for document in res["hits"]["hits"]]

        for _ in range(count // self.batch_length + 1):
            if count <= 0:
                break

            size = count if count <= self.batch_length else self.batch_length
            body = {
                "query": {"match_all": {}},
                "search_after": [res["hits"]["hits"][-1]["_id"]],
                "sort": [{"_id": {"order": "asc"}}],
            }
            res = self.client.search(index=indexName, body=body, size=size)

            result_data.extend(
                [document["_source"] for document in res["hits"]["hits"]]
            )
            count -= self.batch_length

        main_df = pd.DataFrame.from_dict(result_data)
        return main_df

    def run_query(self, indexName, body, **kwargs):
        """Runs search query with respect to given body."""

        return_method = kwargs.get("return_method", None)
        allow_null = kwargs.get("allow_null", False)
        kwargs.pop("return_method", None)
        kwargs.pop("allow_null", None)

        if return_method not in ElasticOperations._implemented_return_methods:
            raise NotImplementedError(
                Fore.RED + f"Return type {return_method} not implemented !!"
            )
        count = self.get_count(indexName)
        if not allow_null and count <= 1:
            raise RuntimeError(
                Fore.RED
                + "Input index is empty!!"
                + " To skip this control set allow_null=True"
            )

        size = kwargs.get("size", count)
        kwargs.pop("size", None)

        try:
            res = self.client.search(
                index=indexName, body=body, size=size, **kwargs
            )
        except RequestError as request_error_message:
            raise RuntimeError(
                Fore.RED + f"query failed with error {request_error_message}"
            ) from request_error_message

        if not allow_null and len(res["hits"]["hits"]) <= 0:
            raise RuntimeError(
                Fore.RED
                + "Result query is empty !!"
                + " To skip this control set allow_null=True"
            )

        if return_method == "hits":
            return res["hits"]

        elif return_method == "total":
            return res["hits"]["total"]

        elif return_method == "value":
            return res["hits"]["total"]["value"]
        elif return_method == "raw":
            return res
        else:
            return res["hits"]["hits"]

    @property
    def ip(self):
        """This function is to get elastic client ip."""
        return self._ip

    @ip.setter
    def ip(self, provided_ip):
        """This function is to set elastic client ip."""
        if provided_ip is not None:
            self._ip = str(provided_ip)
        elif os.environ.get("ELASTIC_CLIENT_IP"):
            self._ip = os.environ.get("ELASTIC_CLIENT_IP")
        else:
            self._ip = None
            print(Fore.RED + "Warning!! Elastic client ip is not defined!")

    @property
    def port(self):
        """This function is to get elastic client port."""
        return self._port

    @port.setter
    def port(self, provided_port):
        """This function is to set elastic client port."""
        if provided_port is not None:
            self._port = int(provided_port)
        elif os.environ.get("ELASTIC_CLIENT_PORT"):
            self._port = os.environ.get("ELASTIC_CLIENT_PORT")
        else:
            self._port = None
            print(Fore.RED + "Warning!! Elastic client port is not defined!")



""" This Module is created to enable Hepsiburada Data Science to communicate with Hive """

import os
import time
from random import randint

import numpy as np
import pandas as pd
from colorama import Fore, init  # #Style

init(autoreset=True)


class HiveOperations:
    """This class is created to enable Hepsiburada Data Science to communicate with Hive"""

    _implemented_returns = ["dataframe", "numpy", "list", "dict"]

    def __init__(self, **kwargs):
        """Initialize the class
        :param HIVE_HOST: Hive Server 2 host
        :param HIVE_PORT: Hive Server 2 port
        :param APP_USER_USERNAME: Hive Server 2 username
        :param APP_USER_PASS: Hive Server 2 password
        """
        self.hive_host = kwargs.get("HIVE_HOST")
        self.hive_port = kwargs.get("HIVE_PORT")
        self.hive_username = kwargs.get("HIVE_USER")
        self.hive_password = kwargs.get("HIVE_PASS")
        self.hadoop_edge_host = kwargs.get("HADOOP_EDGE_HOST")
        self.hive_connection = None

    def __repr__(self):
        """
        Class representation

        Returns:
           str: Returns the class name with connectiton credentials
        """
        return (
            f"HiveOperations(HIVE_HOST={self.hive_host}, HIVE_PORT={self.hive_port}, "
            f'HIVE_USER={self.hive_username}, HIVE_PASS="XXXXXXXX")'
        )

    def connect_to_hive(self):
        """
        This function is used to connect to Hive Server 2 and return the connection object.
        :return: Hive Server 2 connection object
        """
        from pyhive import hive

        try:
            self.hive_connection = hive.connect(
                host=self.hive_host,
                port=self.hive_port,
                username=self.hive_username,
                password=self.hive_password,
                auth="LDAP",
            )
        except TypeError:
            raise TypeError(
                Fore.RED + "Failed to connect to Hive Server 2: "
                "VPN is possibly not connected"
            ) from None

        except ValueError:
            raise ValueError(
                Fore.RED + "Failed to connect to Hive Server 2: "
                "Credentials are incorrect or missing "
            ) from None
        print(Fore.GREEN + "Connection Succeded !!")

    def disconnect_from_hive(self):
        """
        This function is used to disconnect from Hive Server 2.
        :param hive_connection: Hive Server 2 connection object
        :return: None
        """
        self.hive_connection.close()
        print(Fore.GREEN + "Connection Closed! ")

    def execute_query(self, query, return_type="dataframe", **kwargs):
        """
        This function is used to execute the query and return data in different types.
        :param hive_connection: Hive Server 2 connection object
        :param query: Hive query
        :param return_type: Type of return data
        :param kwargs return_colums: Return Columns for numpy array and list types.
        :return: Data frame
        """
        if return_type not in HiveOperations._implemented_returns:
            raise NotImplementedError(
                Fore.RED + f"Return type {return_type} not implemented !!"
            )

        cursor = self.hive_connection.cursor()
        execution_start = time.time()
        cursor.execute(query)
        execution_duration = time.time() - execution_start
        print(
            Fore.YELLOW
            + f"Query executed in {execution_duration:.2f} seconds !"
        )
        columns = HiveOperations._rename_columns(cursor)

        if return_type == "dataframe":
            data = pd.DataFrame(cursor.fetchall(), columns=columns)
        elif return_type == "numpy":
            data = np.array(cursor.fetchall())
        elif return_type == "list":
            data = cursor.fetchall()
        elif return_type == "dict":
            data = [dict(zip(columns, row)) for row in cursor.fetchall()]

        cursor.close()
        if kwargs.get("return_columns"):
            return data, columns
        return data, None

    def create_insert_table(self, query) -> None:
        """
        This function is used to create the table and insert data into the table.
        :param query: Hive query
        :param hive_connection: Hive Server 2 connection object
        :return: None
        """
        cursor = self.hive_connection.cursor()
        cursor.execute(query)
        cursor.close()

    def send_files_to_hive(self, target_table, source_data, columns=None):
        """
        This function is used to send files to Hive Server 2.
        :param target_table: Target Hive table
        :param source_data: Source data
        :param columns: Columns of the source data
        :return: None
        """
        from scp import SCPClient

        unique_filename = str(randint(0, 100000))
        transfer_start = time.time()
        ssh_client = self._set_ssh_client()

        csv_created = False
        edge_node_path = f"/home/{self.hive_username}/Python_CSV"
        hdfs_path = f"/user/{self.hive_username}/projects/Python_CSV"

        # Check if Edge Node & HDFS Path exist
        edge_node_dir_check_cmd = f"ls -d {edge_node_path}"
        hdfs_dir_check_cmd = f"hdfs dfs -test -d {hdfs_path}"

        _, stdout, _ = ssh_client.exec_command(edge_node_dir_check_cmd)
        if stdout.channel.recv_exit_status() != 0:
            print(Fore.RED + f"Edge Node Path {edge_node_path} does not exist")
            print(Fore.YELLOW + f"Creating Edge Node Path {edge_node_path}")
            ssh_client.exec_command(f"mkdir {edge_node_path}")

        _, stdout, _ = ssh_client.exec_command(hdfs_dir_check_cmd)
        if stdout.channel.recv_exit_status() != 0:
            print(Fore.RED + f"HDFS Path {hdfs_path} does not exist")
            print(Fore.YELLOW + f"Creating HDFS Path {hdfs_path}")
            ssh_client.exec_command(f"hdfs dfs -mkdir {hdfs_path}")

        try:
            # Creating a csv file
            if str(type(source_data)) == "<class 'numpy.ndarray'>":
                np.savetxt(
                    unique_filename + ".csv",
                    source_data,
                    delimiter="\t",
                    header="\t".join(columns),
                    comments="",
                )
                source_data = unique_filename + ".csv"
                csv_created = True
                # print("Temporary csv file '" + unique_filename +".csv' created.")

            elif (
                str(type(source_data))
                == "<class 'pandas.core.frame.DataFrame'>"
            ):
                source_data.to_csv(
                    unique_filename + ".csv", index=False, sep="\t"
                )
                source_data = unique_filename + ".csv"
                csv_created = True
                # print("Temporary csv file '" + unique_filename +".csv' created.")

            _, stdout, _ = ssh_client.exec_command(
                f"mkdir {edge_node_path}/{unique_filename};"
            )
            if stdout.channel.recv_exit_status() != 0:
                raise RuntimeError(
                    Fore.RED + "Failed to create directory on edge node"
                )

            with SCPClient(ssh_client.get_transport()) as scp:
                # uploads target CSV to edgenode
                scp.put(
                    source_data,
                    recursive=True,
                    remote_path=f"{edge_node_path}/{unique_filename}/",
                )
            print(Fore.GREEN + "Files sent to EdgeNode succesfully!")

            edge_node_to_hdfs_command_list = [
                f"hadoop fs -rm -r {hdfs_path}/{unique_filename};",
                f"hadoop fs -mkdir {hdfs_path}/{unique_filename}/;",
                f"hadoop fs -put {edge_node_path}/{unique_filename}/*.csv {hdfs_path}/{unique_filename};",
                f"rm -r {edge_node_path}/{unique_filename};",
            ]

            for command in edge_node_to_hdfs_command_list:
                _, stdout, _ = ssh_client.exec_command(command)
                if stdout.channel.recv_exit_status() == 0:
                    print(Fore.GREEN + f"Command executed: {command}")
                else:
                    print(Fore.RED + f"Command failed: {command}")

            print(
                Fore.GREEN
                + "EdgeNode - HDFS transfer completed! \nEdgeNode is clean again!"
            )

            # reading csv file's column types and make column types suitable for hql
            dataframe = pd.read_csv(source_data, sep="\t")
            column_name = dataframe.columns
            dtype_dict = {
                "object": "STRING",
                "int64": "INT",
                "float64": "FLOAT",
                "bool": "BOOLEAN",
            }
            column_type = [dtype_dict[str(item)] for item in dataframe.dtypes]
            columns = []

            for item in zip(column_name, column_type):
                columns.append(" ".join(item))
            column_details = ",".join(columns)

            # numpy,pandas csvleri siliniyor
            if csv_created:
                os.remove(source_data)
                csv_created = False

            # creating external table
            print(Fore.GREEN + "Creating external table...")
            self.create_insert_table(
                f"DROP TABLE IF EXISTS {target_table}_ext"
            )
            self.create_insert_table(
                f"CREATE EXTERNAL TABLE {target_table}_ext"
                f"({column_details}) "
                f'ROW FORMAT DELIMITED FIELDS TERMINATED BY "\t" '
                f'LOCATION "{hdfs_path}/{unique_filename}/" '
                f'TBLPROPERTIES("skip.header.line.count"="1")'
            )

            self.create_insert_table(f"DROP TABLE IF EXISTS {target_table}")
            self.create_insert_table("SET hive.auto.convert.join=false")
            self.create_insert_table("SET hive.tez.container.size=16384")
            self.create_insert_table("SET hive.tez.java.opts=-Xmx13108m")
            self.create_insert_table(
                f"CREATE TABLE {target_table} AS "
                f"SELECT * FROM {target_table}_ext"
            )
            self.create_insert_table(
                f"DROP TABLE IF EXISTS {target_table}_ext"
            )

            print(Fore.GREEN + "Table creation Success!")
            transfer_end = time.time()
            print(
                Fore.YELLOW
                + f"Table Creation took: {transfer_end - transfer_start:.2f} seconds!"
            )

            ssh_client.exec_command(
                f"hadoop fs -rm -r {hdfs_path}/{unique_filename};"
            )
            print(Fore.GREEN + "HDFS is clean again!")

        except Exception as exc:
            ssh_client.exec_command(
                f"hadoop fs -rm -r {hdfs_path}/{unique_filename};"
            )
            print(Fore.GREEN + "HDFS is clean!")
            ssh_client.exec_command(
                f"rm -r {edge_node_path}/{unique_filename};"
            )
            print(Fore.GREEN + "EdgeNode is clean!")
            raise exc

    def _set_ssh_client(self):
        """
        To set ssh client

        Returns:
            SSHClient: returns ssh client object
        """
        from paramiko import AutoAddPolicy, SSHClient

        ssh_client = SSHClient()
        ssh_client.load_system_host_keys()
        ssh_client.set_missing_host_key_policy(AutoAddPolicy())
        ssh_client.connect(
            hostname=self.hadoop_edge_host,
            username=self.hive_username,
            password=self.hive_password,
            look_for_keys=False,
        )
        return ssh_client

    @staticmethod
    def _rename_columns(cursor) -> list:
        """
        This function takes cursor description and rename the columns of the data frame.
        :param column_names: Column names
        :return: List of column names
        """
        columns = [column[0].split(".")[-1] for column in cursor.description]
        return columns

    @staticmethod
    def _parse_sql_query(query):
        """
        This function is used to parse the sql query and return the query with the table name.
        :param query: Hive query
        :return: Hive query with table name
        """

        # TO BE WRITTEN !!

        query_list = query.split(" ")
        table_name = query_list[2]
        return table_name

    @staticmethod
    def _execute_sql_from_file(query_file, hive_connection):
        """
        This function is used to execute the sql query from file and return the data frame.
        :param query_file: Hive query file
        :param hive_connection: Hive Server 2 connection object
        :return: Data frame
        """

        # TO BE WRITTEN !!

        with open(query_file, "r", encoding="utf-8") as hiveql_file:
            query = hiveql_file.read()
        return HiveOperations.execute_query(hive_connection, query)

    @property
    def hive_host(self):
        """Hive Server 2 host"""
        return self._hive_host

    @hive_host.setter
    def hive_host(self, provided_hive_host):
        """Hive Server 2 host setter"""
        if provided_hive_host is not None:
            self._hive_host = str(provided_hive_host)
        elif os.environ.get("HIVE_HOST"):
            self._hive_host = os.environ.get("HIVE_HOST")
        else:
            self._hive_host = None
            print(Fore.RED + "Warning!! Hive host is not specified")

    @property
    def hive_port(self):
        """Hive Server 2 port"""
        return self._hive_port

    @hive_port.setter
    def hive_port(self, provided_hive_port):
        """Hive Server 2 port setter"""
        if provided_hive_port is not None:
            self._hive_port = str(provided_hive_port)
        elif os.environ.get("HIVE_PORT"):
            self._hive_port = os.environ.get("HIVE_PORT")
        else:
            self._hive_port = None
            print(Fore.RED + "Warning!! Hive port is not specified")

    @property
    def hive_username(self):
        """Hive Server 2 username"""
        return self._hive_username

    @hive_username.setter
    def hive_username(self, provided_hive_username):
        """Hive Server 2 username setter"""
        if provided_hive_username is not None:
            self._hive_username = str(provided_hive_username)
        elif os.environ.get("HIVE_USERNAME"):
            self._hive_username = os.environ.get("HIVE_USERNAME")
        else:
            self._hive_username = None
            print(Fore.RED + "Warning!! Hive username is not specified")

    @property
    def hive_password(self):
        """Hive Server 2 password"""
        return self._hive_password

    @hive_password.setter
    def hive_password(self, provided_hive_password):
        """Hive Server 2 password setter"""
        if provided_hive_password is not None:
            self._hive_password = str(provided_hive_password)
        elif os.environ.get("HIVE_PASSWORD"):
            self._hive_password = os.environ.get("HIVE_PASSWORD")
        else:
            self._hive_password = None
            print(Fore.RED + "Warning!! Hive password is not specified")

    @property
    def hadoop_edge_host(self):
        """Hadoop Edge host"""
        return self._hadoop_edge_host

    @hadoop_edge_host.setter
    def hadoop_edge_host(self, provided_hadoop_edge_host):
        """Hadoop Edge host setter"""
        if provided_hadoop_edge_host is not None:
            self._hadoop_edge_host = str(provided_hadoop_edge_host)
        elif os.environ.get("HADOOP_EDGE_HOST"):
            self._hadoop_edge_host = os.environ.get("HADOOP_EDGE_HOST")
        else:
            self._hadoop_edge_host = None
            print(Fore.RED + "Warning!! Hadoop Edge Host is not specified")



"""This module enables Hepsiburada Data Science to communicate with Mongo"""
import time

import pandas as pd
from colorama import Fore, init
from pymongo import MongoClient

init(autoreset=True)


class MongoOperations:
    """This class is created to enable Hepsiburada Data Science to communicate with Mongo"""

    _implemented_returns = ["dataframe", "records", "list", None]

    def __init__(self, **kwargs) -> None:
        self.ip = kwargs.get("ip")
        self.port = kwargs.get("port")
        self.connection_string = kwargs.get("connection_string")
        self.client = self.create_client(
            self.ip, self.port, self.connection_string
        )

    def create_client(
        self, ip: str = None, port: int = None, connection_string: str = None
    ):
        """Creates a client via ip & port or connection string.

        Args:
            ip (str, optional): Ip of Mongo.
            port (int, optional): Port to connect Mongo.
            connection_string (str, optional): Alternative way to connect Mongo

        Returns:
            Mongo Client: Client of the connection.
        """
        if (ip is None or port is None) and connection_string is None:
            raise RuntimeError(Fore.RED + "Mongo client is not defined!!")
        else:
            if connection_string:
                self.client = MongoClient(connection_string)
            else:
                self.client = MongoClient(host=ip, port=port)

        return self.client

    def show_databases(self, in_detail: bool = False):
        """Shows available databases.

        Args:
            in_detail (bool, optional): To show in detail or not.
                                        Defaults to False.
        """
        print("Available Databases:")
        if in_detail:
            databases = self.client.list_databases()
        else:
            databases = self.client.list_database_names()

        for database in databases:
            print(database)

    def show_collections(self, database_name: str, in_detail: bool = False):
        """Shows available collections.

        Args:
            database_name (str): Name of database
            in_detail (bool, optional): To show documents with detailly or not.
                                        Default is False.
        """
        print("Available Collections:")
        database = self.get_database(database_name)
        if in_detail:
            collections = database.list_collections()
        else:
            collections = database.list_collection_names()

        for collection in collections:
            print(collection)

    def create_database(self, database_name: str):
        """Creates database database with respect to input name.
           Creates database if it does not exist.

        Args:
            database_name (str): Database name to be created.

        Returns:
            Mongo Database: Group of Mongo to store collections.
        """
        if database_name not in self.client.list_database_names():
            database = self.client[database_name]
            database.create_collection("default")
            print(Fore.GREEN + f"{database_name} named database is created.")

        else:
            database = self.client[database_name]
            print(
                Fore.YELLOW + f"{database_name} named database already exist."
            )

        return database

    def create_collection(
        self, database_name: str, collection_name: str, **kwargs
    ):
        """Gets collection with respect to input name.
           Creates collection if it does not exist.

        Args:
            database_name (str): Database name as a parent of collection.
            collection_name (str): Collection name to be created.

        Returns:
            Mongo collection: Group of Mongo to store documents.
        """

        database = self.get_database(database_name)
        if collection_name not in database.list_collection_names():
            collection = database.create_collection(collection_name, **kwargs)
            print(
                Fore.GREEN + f"{collection_name} named collection is created."
            )
        else:
            collection = database[collection_name]
            print(
                Fore.YELLOW
                + f"{collection_name} named collection already exist."
            )

        return collection

    def get_database(self, database_name: str):
        """Gets database database with respect to input name.

        Args:
            database_name (str): Database name to be obtained.

        Returns:
            Mongo Database: Group of Mongo to store collections.
        """

        if database_name not in self.client.list_database_names():
            raise RuntimeError(
                Fore.RED
                + f"{database_name} does not exist. Call create_dabatase() function to create it."
            )
        else:
            database = self.client[database_name]
            return database

    def get_collection(self, database_name: str, collection_name: str):
        """Gets collection with respect to input name.

        Args:
            database_name (str): Database name as a parent of collection.
            collection_name (str): Collection name to be created.

        Returns:
            Mongo collection: Group of Mongo to store documents.
        """

        database = self.get_database(database_name)
        if collection_name not in database.list_collection_names():
            raise RuntimeError(
                Fore.RED
                + f"{collection_name} does not exist. Call create_collection() function to create it."
            )
        else:
            collection = database[collection_name]
            return collection

    def get_count(self, database_name: str, collection_name: str, filter={}):
        """Counts the document number in that collection.

        Args:
            database_name (str): Name of the database
            collection_name (str): Name of the collection
            filter (dict, optional): Filters documents that will be counted.
                                     Default is {}.
        Returns:
            int: Returns the document count of the collection.
        """
        collection = self.get_collection(database_name, collection_name)
        return collection.count_documents(filter=filter)

    def find_document(
        self, database_name: str, collection_name: str, **kwargs
    ):
        """Finds document from the collection.

        Args:
            database_name (str): Name of database
            collection_name (str): Name of collection
        Returns:
            Mongo document: Set of field-value pairs.
        """
        collection = self.get_collection(database_name, collection_name)

        filter = kwargs.get("filter")
        projection = kwargs.get("projection")
        sort = kwargs.get("sort")
        # 0 prevent limiting
        limit = kwargs.get("limit", 0)
        # List of dictionaries
        return_type = kwargs.get("return_type", "records")
        allow_null = kwargs.get("allow_null", False)
        if return_type not in MongoOperations._implemented_returns:
            raise NotImplementedError(
                Fore.RED + f"Return type {return_type} not implemented !!"
            )

        if type(sort) == tuple:
            sort = [sort]
        try:
            result = collection.find(
                filter=filter, sort=sort, limit=limit, projection=projection
            )
        except (ValueError, TypeError) as request_error_message:
            raise RuntimeError(
                Fore.RED + f"Command failed with error {request_error_message}"
            ) from request_error_message

        documents = [doc for doc in result]

        if not allow_null and len(documents) == 0:
            raise RuntimeError(
                Fore.RED
                + "Obtained data is empty!! To skip this control set allow_null=True "
            )

        if return_type == "dataframe":
            return pd.DataFrame.from_dict(documents)
        return documents

    def insert_document(self, database_name: str, collection_name: str, data):
        """Inserts document to the collection.

        Args:
            database_name (str): Name of database
            collection_name (str): Name of collection
            data (records or dict): To insert one data only one dict should be passed.
                                 To insert many data list of dicts should be passed.
        """

        collection = self.get_collection(database_name, collection_name)

        if type(data) == pd.DataFrame:
            data = data.to_dict("records")
        elif type(data) == dict:
            data = [data]

        if len(data) == 0:
            print(Fore.RED + "Input data is empty!!")

        if type(data) == list:
            try:
                result = collection.insert_many(data)
            except (ValueError, TypeError) as request_error_message:
                raise RuntimeError(
                    Fore.RED
                    + f"Command failed with error {request_error_message}"
                ) from request_error_message

            count = len(result.inserted_ids)
            print(Fore.GREEN + f"{count} documents inserted.")
        else:
            print(
                Fore.RED
                + "Could not insert data because type "
                + f"{type(data)} is not implemented!!"
            )
        # Time was required to insert document into cluster
        time.sleep(0.01)

    def update_document(
        self,
        database_name: str,
        collection_name: str,
        filter: dict,
        new_values: dict,
    ):
        """Updates document with respect to filter and new values.

        Args:
            database_name (str): Name of database
            collection_name (str): Name of collection
            filter (dict): Condition to select documents
            new_values (dict): New values to update the document.
        """
        collection = self.get_collection(database_name, collection_name)
        try:
            result = collection.update_many(filter, new_values)
        except (ValueError, TypeError) as request_error_message:
            raise RuntimeError(
                Fore.RED + f"Command failed with error {request_error_message}"
            ) from request_error_message

        count = result.modified_count
        if count == 0:
            print(Fore.YELLOW + "No documents updated!!")

        else:
            print(Fore.GREEN + f"{count} documents updated.")
        time.sleep(0.01)

    def delete_document(
        self, database_name: str, collection_name: str, filter: dict
    ):
        """Deletes document with respect to filter.

        Args:
            database_name (str): Name of database
            collection_name (str): Name of collection
            filter (dict): Filter to select documents to be deleted.
        """
        collection = self.get_collection(database_name, collection_name)
        try:
            count = collection.delete_many(filter).deleted_count
        except (ValueError, TypeError) as request_error_message:
            raise RuntimeError(
                Fore.RED + f"Command failed with error {request_error_message}"
            ) from request_error_message

        if count == 0:
            print(Fore.YELLOW + "No documents deleted!!")
        else:
            print(Fore.GREEN + f"{count} documents deleted.")
        time.sleep(0.01)

    def drop_database(self, database_name: str):
        """Drops the input database.

        Args:
            database (str): Name of database to be dropped.
        """
        if database_name not in self.client.list_database_names():
            print(
                Fore.YELLOW
                + f"{database_name} named database "
                + "has no collection or does not exist."
            )
        else:
            self.client.drop_database(database_name)
            print(Fore.GREEN + f"{database_name} named database is deleted.")

    def drop_collection(self, database_name: str, collection_name: str):
        """Drops the input collection

        Args:
            database (str): Group of Mongo collections.
            collection_name (str): Name of collection to be dropped.
        """
        database = self.get_database(database_name)
        if collection_name not in database.list_collection_names():
            print(
                Fore.YELLOW
                + f"{collection_name} named collection "
                + "already does not exist."
            )
        else:
            database.drop_collection(collection_name)
            print(
                Fore.GREEN + f"{collection_name} named collection is deleted."
            )

    @property
    def ip(self):
        """This function is to get Mongo client ip."""
        return self._ip

    @ip.setter
    def ip(self, provided_ip):
        """This function is to set Mongo client ip."""
        self._ip = provided_ip

    @property
    def port(self):
        """This function is to get Mongo client port."""
        return self._port

    @port.setter
    def port(self, provided_port):
        """This function is to set Mongo client port."""
        self._port = provided_port

    @property
    def connection_string(self):
        """This function is to get Mongo client connection string."""
        return self._connection_string

    @connection_string.setter
    def connection_string(self, provided_string):
        """This function is to set Mongo client connection string."""
        self._connection_string = provided_string



""" This Module is created to enable Hepsiburada Data Science to communicate with Cloud Storage. """

import json
import os
import pickle
import shutil
import zipfile
from io import BytesIO

import pandas as pd
from colorama import Fore, init  # #Style
from google.api_core.exceptions import NotFound
from google.cloud import storage
from google.oauth2 import service_account

init(autoreset=True)


class BucketOperations(storage.Client):
    """This class is created to enable Hepsiburada Data Science to communicate with Cloud Storage"""

    _implemented_type = ["dataframe", "json", "pickle"]

    def __init__(self, **kwargs) -> None:
        self.gcp_key = kwargs.get("gcp_key_path")
        self.credentials = kwargs.get("gcp_key_path")
        super().__init__(
            project=self.credentials.project_id, credentials=self.credentials
        )

    def __repr__(self) -> str:
        """This function returns project id and service account email."""
        return (
            f"{self.__class__.__name__}(Project ID: {self.credentials.project_id}, "
            f"Service Account: {self.credentials._service_account_email.split('@')[0]})"
        )

    def does_blob_exist(self, bucket_name: str, blob_name: str) -> bool:
        """_summary_

        Args:
            bucket_name (str): Bucket name
            blob_name (str): Blob name to be checked

        Returns:
            bool: Returns True if blob exists, otherwise False
        """
        bucket = self.bucket(bucket_name)
        blob = bucket.blob(blob_name)

        return blob.exists()

    def upload_from_filepath(self, bucket_name, blob_name, filepath):
        """This function is to upload data from filepath."""
        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.upload_from_filename(filepath)

            if not blob.exists():
                raise RuntimeError(Fore.RED + f"Could not upload {blob_name}.")
            file_size = os.stat(filepath).st_size
            if file_size != blob.size:
                raise RuntimeError(
                    Fore.RED
                    + f"Size of {blob_name} is {blob.size},"
                    + f"but the size of {filepath} is {file_size} "
                )

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

        except Exception:
            print(f"No such file in this path: <{filepath}>")

    def upload_from_memory(
        self,
        bucket_name,
        blob_name,
        contents,
        upload_type,
        check_upload: bool = True,
    ):
        """This function is to upload data from memory."""
        if upload_type not in BucketOperations._implemented_type:
            raise NotImplementedError(
                Fore.RED + f"Download type {upload_type} not implemented !!"
            )

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)

            if upload_type == "json":
                send_data = json.dumps(contents, indent=4).encode("utf-8")
            elif upload_type == "pickle":
                send_data = pickle.dumps(contents)
            elif upload_type == "dataframe":
                send_data = contents.to_csv()

            blob.upload_from_string(send_data)

            if check_upload and not blob.exists():
                raise RuntimeError(f"{blob_name} does not exist.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def download_to_filepath(
        self, bucket_name, blob_name, filepath, extract=False
    ):
        """This function is to download data from filepath."""
        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            if blob.exists():
                with open(filepath, "wb") as f:
                    self.download_blob_to_file(blob, f)

                if extract:
                    file_name = filepath.split("/")[-1]
                    folder_path = filepath[: -len(file_name)]
                    shutil.unpack_archive(filepath, folder_path, "zip")

            else:
                print(f"Blob <{blob_name}> does not exist.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

        except Exception as ex:
            print(ex)
            print(f"No such file in this path: <{filepath}>")

    def download_to_memory(self, bucket_name, blob_name, download_type):
        """This function is to download data from memory."""
        if download_type not in BucketOperations._implemented_type:
            raise NotImplementedError(
                Fore.RED + f"Download type {download_type} not implemented !!"
            )

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            if blob.exists():
                if download_type == "dataframe":
                    df = pd.read_csv(
                        BytesIO(blob.download_as_string()),
                        encoding="UTF-8",
                        sep=",",
                    )
                    return df

                elif download_type == "json":
                    data = (
                        blob.download_as_string()
                        .decode("utf8")
                        .replace("'", '"')
                    )
                    load_json = json.loads(data)
                    json_data = json.dumps(load_json, indent=4, sort_keys=True)
                    return json_data

                elif download_type == "pickle":
                    data = blob.download_as_bytes()
                    try:
                        load_pickle = pickle.loads(data)
                    except pickle.UnpicklingError:
                        load_pickle = BytesIO(data)

                    return load_pickle
            else:
                print(f"Blob <{blob_name}> does not exist.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def create_new_folders(self, bucket_name, folder_name):
        """This function is to create new folder from the bucket."""
        try:
            bucket = self.bucket(bucket_name)
            dst_bucket = self.bucket(bucket_name)

            blob = bucket.blob("New_Folder_Starter/")
            new_blob = bucket.copy_blob(blob, dst_bucket, new_name=folder_name)

            new_blob.acl.save(blob.acl)
            print(f"<{folder_name}> folder created.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def delete_folder_from_bucket(self, bucket_name, folder_name):
        """This function is to create new folder from the bucket."""
        try:
            bucket = self.get_bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=folder_name)
            for blob in blobs:
                blob.delete()

            print(f"Folder <{folder_name}> deleted.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def delete_file_from_bucket(self, bucket_name, blob_name):
        """This function is to delete file from the bucket."""

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.delete()
            print(f"Blob <{blob_name}> deleted.")

        except NotFound:
            print(f"No such file in this path: <{bucket_name}/{blob_name}>")

    def download_recursively(
        self,
        bucket_name: str,
        blob_name: str,
        overwrite: bool,
        folder_path: str,
        as_zip: bool = False,
    ):
        """Downloads all files from given bucket directory.

        Args:
            bucket_name (str): Name of bucket
            blob_name (str): Parent directory of files at bucket.
            overwrite (bool): Determines to write over existing files or not.
            folder_path (str): Parent directory of location to be downloaded at local.
        """
        blob_name = blob_name + "/" if blob_name[-1] != "/" else blob_name

        blobs = self.list_blobs(bucket_name, prefix=blob_name)
        if as_zip:
            os.makedirs(folder_path, exist_ok=True)
            bucket = self.bucket(bucket_name)
            blob_names = [blob.name for blob in blobs]

            zip_name = blob_name.split("/")[-2] + ".zip"
            download_path = os.path.join(folder_path, zip_name)
            with zipfile.ZipFile(download_path, "w") as zip:
                for sub_blob_name in blob_names:
                    blob = bucket.blob(sub_blob_name)
                    data = blob.download_as_string()
                    sub_file_path = sub_blob_name[len(blob_name) :]
                    if len(sub_file_path) == 0:
                        continue
                    zip.writestr(sub_file_path, data)
            return

        for blob in blobs:
            download_path = os.path.join(
                folder_path, blob.name[len(blob_name) :]
            )
            if len(download_path) == 0:
                continue
            idx = download_path.rfind("/")
            if not os.path.exists(download_path[:idx]) and (
                idx != -1 or download_path[-1] == "/"
            ):
                os.makedirs(download_path[:idx], exist_ok=True)

            if download_path[-1] == "/":
                continue
            if not os.path.exists(download_path) or (
                os.path.exists(download_path) and overwrite
            ):
                self.download_to_filepath(
                    bucket_name, blob.name, download_path
                )

    def upload_recursively(
        self,
        bucket_name: str,
        blob_name: str,
        overwrite: bool,
        folder_path: str,
        as_zip=False,
    ):
        """Downloads all files from given local directory.

        Args:
            bucket_name (str): Name of bucket
            blob_name (str): Parent directory of files to be uploaded at bucket.
            overwrite (bool): Determines to write over existing files or not.
            folder_path (str): Parent directory of source files at local.
            as_zip (bool, optional): Upload a folder as zip to bucket.
                                     Default is False.
        """
        folder_path = (
            folder_path + "/" if folder_path[-1] != "/" else folder_path
        )
        blob_name = blob_name + "/" if blob_name[-1] != "/" else blob_name

        blobs = self.list_blobs(bucket_name, prefix=blob_name)
        blob_names = [blob.name for blob in blobs]

        if as_zip:
            zip_name = folder_path.split("/")[-2]
            zip_path = os.path.join(
                folder_path[: -(len(zip_name) + 1)], zip_name
            )
            zip_name += ".zip"
            upload_file_path = blob_name + zip_name
            source_file_path = zip_path + ".zip"
            if upload_file_path not in blob_names or (
                upload_file_path in blob_names and overwrite
            ):
                shutil.make_archive(zip_path, "zip", folder_path)
                self.upload_from_filepath(
                    bucket_name, upload_file_path, source_file_path
                )
            return

        for path, dirs, files in os.walk(folder_path):
            for dir in dirs:
                upload_dir_path = os.path.join(
                    blob_name, path[len(folder_path) :], dir + "/"
                )
                if upload_dir_path not in blob_names:
                    self.create_new_folders(bucket_name, upload_dir_path)

            for file in files:
                upload_file_path = os.path.join(
                    blob_name, path[len(folder_path) :], file
                )
                source_file_path = os.path.join(path, file)
                if upload_file_path not in blob_names or (
                    upload_file_path in blob_names and overwrite
                ):
                    self.upload_from_filepath(
                        bucket_name, upload_file_path, source_file_path
                    )

    @property
    def gcp_key(self):
        """This function is to get GCP key."""
        return self._gcp_key

    @gcp_key.setter
    def gcp_key(self, provided_gcp_key):
        """This function is to set GCP key."""
        if provided_gcp_key is not None:
            self._gcp_key = str(provided_gcp_key)
            self.credentials = str(provided_gcp_key)
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._gcp_key = os.environ.get("SERVICE_ACCOUNT_KEY_PATH")
        else:
            self._gcp_key = None
            print(
                Fore.RED
                + "Warning!! GCP Key Path for Service Account is not specified"
            )

    @property
    def credentials(self):
        """This function is to get credentials."""
        return self._credentials

    @credentials.setter
    def credentials(self, provided_credentials):
        """This function is to set credentials."""
        if provided_credentials is not None:
            self._credentials = (
                service_account.Credentials.from_service_account_file(
                    self.gcp_key,
                    scopes=["https://www.googleapis.com/auth/cloud-platform"],
                )
            )
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._credentials = (
                service_account.Credentials.from_service_account_file(
                    self.gcp_key,
                    scopes=["https://www.googleapis.com/auth/cloud-platform"],
                )
            )
        else:
            self._credentials = None
            print(
                Fore.RED
                + "Warning!! Credentials for Service Account is not specified"
            )
